$DebugTraj = false;

if($Pref::staticHitscan::stepSpeed $= "")
	$Pref::staticHitscan::stepSpeed = 33;
	
$WindVector = "0 0 0";

datablock StaticShapeData(BulletStaticShape)
{
	shapeFile = "./bullet.dts";
};

function getBounceVector(%vector, %normal)
{
	%v = %vector;
	%n = %normal;
	return vectorSub(%v, vectorScale(%n, 2 * vectorDot(%v, %n)));
}
function getStepDistance(%data)
{
	%ups = %data.unitsPerSecond;
	
	%updates = 1000 / $Pref::staticHitscan::stepSpeed;
	%distance = %ups / %updates;
	
	return %distance;
}
function getSwayVector(%sway)
{
	return getSway(%sway) SPC getSway(%sway) SPC 0;
}
function getSway(%sway)
{
	return getRandom() * 0.0001 * getRandom(-1, 1) * %sway;
}
function createStaticHitscan_rWepPack(%player, %initialPosition, %vector, %projectile, %spread, %zero, %stepSpeed, %stepDistance, %step, %initPos, %staticObject)
{
	// First step
	
	if(!%step)
	{
		%initPos = %initialPosition;
		
		// zero
		%vector = vectorNormalize(vectorAdd(%vector, "0 0" SPC %zero));
		
		// spread 
		if(%spread > 0)
		{
			%x = (getRandom() - 0.5) * 10 * 3.1415926 * (%spread);
			%y = (getRandom() - 0.5) * 10 * 3.1415926 * (%spread);
			%z = (getRandom() - 0.5) * 10 * 3.1415926 * (%spread);
			%mat = MatrixCreateFromEuler(%x @ " " @ %y @ " " @ %z);
			%vector = MatrixMulVector(%mat, %vector);
		}
		
		//$SHP_StartTime = getSimTime();
	}
	
	//$SHP_CurrentTime = getSimTime();
	
	// player vars
	%client = %player.client;
	
	// projectile vars
	%data = %projectile;
	%stepDistance = getStepDistance(%data);
	%stepSpeed = $Pref::staticHitscan::stepSpeed;
	%ricochet = %data.allowRicochet;
		
	// init raycast
	%start = %initialPosition;
	%end = vectorAdd(%start, vectorScale(%vector, %stepDistance));
	%mask = $TypeMasks::FxBrickObjectType |
			$TypeMasks::PlayerObjectType |
			$TypeMasks::VehicleObjectType |
			$TypeMasks::TerrainObjectType;
						
	%raycast = containerRayCast(%start, %end, %mask, %player);
	%hit = getWord(%raycast, 0);
	%hitPosition = getWords(%raycast, 1, 3);
	
	// for when we hit something
	if(%hit)
	{
		if(isObject(%staticObject))
			%staticObject.delete();
	
		// create the projectile at the location of the impact
		%realProjectile = new Projectile()
		{
			datablock = %projectile;
			initialPosition = vectorAdd(getWords(%raycast, 1, 3), vectorScale(%vector, -0.1));
			initialVelocity = vectorScale(%vector, 200);
			client = %client;
			sourceObj = %player;
			sourceObject = %player;
			minigame = %minigame;
		};
		
		// in some cases with this method the shooter can shoot themselves (???) so prevent this from causing damage
		%realProjectile.dontHurtShooter = true;
		return;
	}
	
	// didn't hit anything
	// get some info about our projectile
	%effectiveRange = %data.effectiveRange;
	%totalRange = %data.totalRange;
	%gravity = %data.gravityScale;
	%sway = %data.swayMod;
	%speedBonus = %data.effectiveSpeedBonus;
	%useFakeProjectileEffect = %data.spawnFakeProjectiles;
	%useTracer = %data.tracerEffect;
	%scaleCalibre = %data.scaleCalibre;
	%scaleLength = %data.scaleLength;
	
	// didn't find what we needed, set def values
	if(%totalRange $= "")
		%totalRange = 2000;
		
	if(%effectiveRange $= "")
		%effectiveRange = 200;
		
	if(%sway $= "")
		%sway = 0;
		
	if(%speedBonus $= "")
		%speedBonus = 1;
		
	// get our current range
	%currentRange = vectorDist(%initPos, %end);
	
	if(%currentRange > %effectiveRange)
	{
		// outside effective range
		// wind and sway
		%vector = vectorNormalize(vectorAdd(vectorAdd(%vector, vectorScale($WindVector, %sway)), getSwayVector(%sway)));
		
		// gravity
		%gravity = %gravity * 0.001;
		%vector = vectorNormalize(vectorAdd(%vector, "0 0" SPC %gravity * -1));
	}
	
	// create a static shape that gives the illusion of a bullet traveling very fast, move it to a new position each time
	if(%useFakeProjectileEffect)
	{
		// aligning static shape to direction of travel
		// written by conan
		%rot = vectorNormalize(%vector);
		%xyz = vectorNormalize(vectorCross("1 0 0", %rot)); //rotation axis
		%u = mACos(vectorDot("1 0 0", %rot)) * -1; //rotation value
		%rot = %xyz SPC %u;

		// if we already have a static shape, just move it to the new position
		if(isObject(%staticObject))
		{
			%staticObject.setTransform(%start SPC %rot);
			%staticObject.playThread(1, travel);
		}
		else
		{
			// if not, create a new one
			if(%step > 0)
			{	
				//if(%useTracer)
				//	%datablock = TracerStaticShape;
				//else
				%datablock = BulletStaticShape;
				
				
				%staticObject = new StaticShape()
				{
					datablock = %datablock;
				};
				%staticObject.setTransform(%start SPC %rot);
				%staticObject.setScale(%scaleLength SPC %scaleCalibre SPC %scaleCalibre);
				%staticObject.playThread(1, travel);
			}
		}
	}
	
	if(isFunction(%data.getName(), "onStaticHitscanStep"))
	{
		%data.onStaticHitscanStep(%player, %start, %end, %vector, %staticObject);
	}

	// we're outside of our total range, don't continue
	if(%currentRange > %totalRange)
	{
		if(isObject(%staticObject))
			%staticObject.delete();
			
		return;
	}
	
	// sentinel for bogus loops, 1000 is still even a massive value, could probably be 200
	if(%step > 1000)
	{
		if(isObject(%staticObject))
			%staticObject.delete();
	
		return;
	}
	
	// next step
	schedule(%stepSpeed, 0, createStaticHitscan_rWepPack, %player, %end, %vector, %projectile, %spread, %zero, %stepSpeed, %stepDistance, %step++, %initPos, %staticObject);
}

package staticProjectilePackage
{
	function ProjectileData::onCollision(%this, %obj, %col, %fade, %pos, %normal, %velocity)
	{		
		// Prevents the user from shooting themselves when shooting directly down or against a wall
		if(%obj.dontHurtShooter && %col.client == %obj.client && %col.getClassName() $= "Player")
		{
			return;
		}
		
		Parent::onCollision(%this, %obj, %col, %fade, %pos, %normal, %velocity);
	
		// Ricochet functionality
		if(%this.allowRicochet)
		{		
			%reflectVec = getBounceVector(%velocity, %normal);
			%dot = vectorDot(%velocity, %reflectVec);
			%dot = %dot / (vectorLen(%velocity) * vectorLen(%reflectVec));
			%acos = mACos(%dot);
			%angle = mRadtoDeg(%acos);
			
			//// vars
			%ricoMinimumAngle = %this.ricoMinimumAngle;
			%ricoChanceCurve = %this.ricoChanceCurve;
			%ricoRarity = %this.ricoRarity;
			
			if(%angle > %ricoMinimumAngle)
				return;
			
			%chanceMax = mCeil((%angle / %ricoChanceCurve) + %ricoRarity); // rico occurs very often at obtuse angles, and almost never towards right angle. minimum %50 chance at any angle
			%chance = getRandom(0, %chanceMax);
				
			if(!%chance)
			{
				//// Ricochet sound
				serverplay3d("fastRicoSound" @ getRandom(1, 2), %obj.getPosition() SPC "0 0 0 0");
				
				// Look for custom functionality. If we don't find any, but the user still wants it to ricochet,
				// just re-do the hitscan using the bounce vector.
				if(!isFunction(%this.getName(), "onRicochet"))
				{
					createStaticHitscan(%obj.client, %pos, %reflectVec, %this, 0.0001);
				}
				else
				{
					%this.onRicochet(%obj, %pos, %reflectVec);
				}
			}
		}
	}
};
activatePackage(staticProjectilePackage);